/*******************************************************************************************/
/**                                                                                       **/
/** COPYRIGHT (C) 2010, Systemyde International Corporation, ALL RIGHTS RESERVED          **/
/**                                                                                       **/
/** serial transmitter module (async/sync)                            Rev 0.0  07/21/2010 **/
/**                                                                                       **/
/*******************************************************************************************/
module stx_byte (bufr_done, bufr_empty, serx_tdat, tx_run, auto_trig, cks_mode, clkp,
                 ld_sxdr, pwrite_bus, s_reset, tx_sync);

  input         cks_mode;      /* clocked serial mode                                      */
  input         clkp;          /* main peripheral clock                                    */
  input         ld_sxdr;       /* data register write                                      */
  input         s_reset;       /* synchronous reset                                        */
  input         tx_sync;       /* transmitter clock enable                                 */
  input   [7:0] pwrite_bus;    /* internal peripheral write bus                            */
  output        auto_trig;     /* clocked serial trigger                                   */
  output        bufr_done;     /* tranmsitter idle                                         */
  output        bufr_empty;    /* transmitter buffer empty                                 */
  output        serx_tdat;     /* transmitter raw output                                   */
  output        tx_run;        /* transmitter is running                                   */

  /*****************************************************************************************/
  /*                                                                                       */
  /* signal declarations                                                                   */
  /*                                                                                       */
  /*****************************************************************************************/
  wire        auto_trig;                                   /* clocked serial trigger       */
  wire        bitclk_pls;                                  /* tx bit clock                 */
  wire        bufr_done;                                   /* async byte finished          */
  wire        bufr_empty;                                  /* buffer empty output          */
  wire        divide_pls;                                  /* bit cell boundary/middle     */
  wire        lastbit;                                     /* shifter empty                */
  wire        ld_bufr;                                     /* load sr with data            */
  wire        ld_shft;                                     /* shifter load control         */
  wire        serx_tdat;                                   /* transmitter raw output       */
  wire        termchar;                                    /* terminate char tx            */
  wire        tx_run;                                      /* cks running                  */
  wire  [7:0] bufr_rev;                                    /* msb-first data               */

  reg         bitclk_reg;                                  /* bit clock                    */
  reg         bufr_full;                                   /* buffer full flag             */
  reg         ending_reg;                                  /* stop bit sending             */
  reg         output_reg;                                  /* output register              */
  reg         running_reg;                                 /* sending byte                 */
  reg         strt_reg;                                    /* tx start cmd latched         */
  reg   [3:0] divider_nxt, divider_reg;                    /* clock divider                */
  reg   [7:0] bufr_reg;                                    /* data buffer byte             */
  reg   [9:0] shft_nxt, shft_reg;                          /* tx shifter                   */

  /*****************************************************************************************/
  /*                                                                                       */
  /* serial transmitter control signals                                                    */
  /*                                                                                       */
  /*****************************************************************************************/
  assign auto_trig   = cks_mode && ld_sxdr;
  assign bitclk_pls  = tx_sync && ((divide_pls && !bitclk_reg) || cks_mode);
  assign bufr_done   = !bufr_full && !running_reg;
  assign bufr_empty  = !bufr_full;
  assign bufr_rev    = {bufr_reg[0], bufr_reg[1], bufr_reg[2], bufr_reg[3],
                        bufr_reg[4], bufr_reg[5], bufr_reg[6], bufr_reg[7]};
  assign divide_pls  = divider_reg[3] && !divider_reg[2];
  assign lastbit     = ~|shft_reg[9:3] && ((cks_mode && shft_reg[2]) ||
                                          (!cks_mode && ~|shft_reg[2:1] && shft_reg[0]));
  assign ld_bufr     = bufr_full &&
                       ((tx_sync && !running_reg && (!cks_mode || strt_reg)) ||
                       (!cks_mode && ending_reg && bitclk_pls));
  assign ld_shft     = bitclk_pls && running_reg;
  assign serx_tdat   = (cks_mode) ? !shft_reg[0] : !output_reg;
  assign termchar    = (cks_mode) ? lastbit : (ending_reg && bitclk_pls);
  assign tx_run      = cks_mode && (strt_reg || running_reg || (ending_reg && !tx_sync));

  always @ (cks_mode or ld_bufr or bufr_reg or bufr_rev or shft_reg) begin
    case ({cks_mode, ld_bufr})
      2'b01:   shft_nxt = {1'b1, 1'b0,        !bufr_reg[7], ~bufr_reg[6:0]};
      2'b11:   shft_nxt = {1'b0, 1'b1,        !bufr_rev[7], ~bufr_rev[6:0]};
      default: shft_nxt = {1'b0, shft_reg[9], shft_reg[8],   shft_reg[7:1]};
      endcase
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* serial transmitter buffer and status                                                  */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (posedge clkp) begin
    bufr_full <= !s_reset && (ld_sxdr  || (!ld_bufr && bufr_full));
    end

  always @ (posedge clkp) begin
    if (ld_sxdr) bufr_reg <= pwrite_bus;
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* serial transmitter shifter                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (posedge clkp) begin
    if (ld_bufr || ld_shft || s_reset) begin
      output_reg <= !s_reset && (ld_bufr || (lastbit ^ shft_reg[0]));
      shft_reg   <= (s_reset) ? 10'h000 : shft_nxt;
      end
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* serial transmitter state machine                                                      */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (divider_reg or running_reg) begin
    casex ({running_reg, divider_reg})
      5'b0xxxx,
      5'b110xx: divider_nxt = 4'b0000;
      5'b10000: divider_nxt = 4'b0001;
      default:  divider_nxt = {divider_reg[2:0], !divider_reg[3]};
      endcase
    end

  always @ (posedge clkp) begin
    if (tx_sync || s_reset) begin
      running_reg <= !s_reset && (ld_bufr ||
                                 (running_reg && ((!cks_mode && bufr_full) || !termchar)));
      divider_reg <= (s_reset) ? 4'h0 : divider_nxt;
      bitclk_reg  <= !s_reset && (ld_bufr || (running_reg && (divide_pls ^ bitclk_reg)));
      end
    end

  always @ (posedge clkp) begin
    if (bitclk_pls || s_reset) ending_reg <= !s_reset && lastbit;
    end

  always @ (posedge clkp) begin
    strt_reg <= !s_reset && (auto_trig || (!ld_bufr && strt_reg));
    end

  endmodule






